
local min, max, floor, abs, sqrt = math.min, math.max, math.floor, math.abs, math.sqrt
local sin, cos, rad, pi, half_pi = math.sin, math.cos, math.rad, math.pi, math.pi / 2
local table_sort = table.sort

-- HELPERS

local function normalized(x, y)
    local l = sqrt(x * x + y * y)
    return x / l, y / l
end

local function sq_dist(x, y, x0, y0)
    x, y = x - x0, y - y0
    return x * x + y * y
end

local function segmentVsSegment(x1, y1, x2, y2, x3, y3, x4, y4)
    local dx1, dy1 = x2 - x1, y2 - y1
    local dx2, dy2 = x4 - x3, y4 - y3
    local dx3, dy3 = x1 - x3, y1 - y3
    local d = dx1 * dy2 - dy1 * dx2
    
    if d == 0 then return false end
    
    local t1 = (dx2 * dy3 - dy2 * dx3) / d
    if t1 < 0 or t1 > 1 then return false end
    
    local t2 = (dx1 * dy3 - dy1 * dx3) / d
    if t2 < 0 or t2 > 1 then return false end
    
    -- point of intersection
    return true, x1 + t1 * dx1, y1 + t1 * dy1
end

local function rotate(x, y, r, cx, cy)
    local c, s = cos(r), sin(r)
    x, y = x - cx, y - cy
    
    return cx + x * c - y * s, cy + x * s + y * c
end

local function online(x, y, ax, ay, bx, by)
    x = ((bx - ax) * (y - ay) - (by - ay) * (x - ax))
    return x > 0 and 1 or (x < 0 and -1 or 0)
end

local function side(x11, y11,   x12, y12,   x21, y21,   x22, y22)
	local z1 = (x22 - x21) * (y11 - y21) - (x11 - x21) * (y22 - y21)
	local z2 = (x22 - x21) * (y12 - y21) - (x12 - x21) * (y22 - y21)
	return z1 * z2
end

local function ccw(px, py, qx, qy, rx, ry)
    local x1, y1, x2, y2 = qx - px, qy - py, rx - px, ry - py
    return x1 * y2 - y1 * x2 >= 0
end

local function sort_up(a, b)
    if a.x == b.x then
        return a.y < b.y
    else
        return a.x < b.x 
    end
end

local function sort_down(a, b)
    if a.x == b.x then
        return a.y > b.y
    else
        return a.x > b.x 
    end
end

local function sort_d(a, b)
    return a.d < b.d 
end

-- VERTICES 

local function v_box(v)
    local l, t, r, b = min(v[1].x, v[2].x, v[3].x, v[4].x), min(v[1].y, v[2].y, v[3].y, v[4].y), max(v[1].x, v[2].x, v[3].x, v[4].x), max(v[1].y, v[2].y, v[3].y, v[4].y)
    return l, t, r, b, r - l, b - t
end

local function v_center(v)
    return 0.25 * (v[1].x + v[2].x + v[3].x + v[4].x), 0.25 * (v[1].y + v[2].y + v[3].y + v[4].y)
end

local function v_move(v, dx, dy)
    v[1].x, v[1].y = v[1].x + dx, v[1].y + dy
    v[2].x, v[2].y = v[2].x + dx, v[2].y + dy
    v[3].x, v[3].y = v[3].x + dx, v[3].y + dy
    v[4].x, v[4].y = v[4].x + dx, v[4].y + dy
end

local function v_rotate(v, x, y, r)
    local c, s = cos(r), sin(r)
    v[1].x, v[1].y = x + (v[1].x - x) * c - (v[1].y - y) * s, y + (v[1].x - x) * s + (v[1].y - y) * c 
    v[2].x, v[2].y = x + (v[2].x - x) * c - (v[2].y - y) * s, y + (v[2].x - x) * s + (v[2].y - y) * c 
    v[3].x, v[3].y = x + (v[3].x - x) * c - (v[3].y - y) * s, y + (v[3].x - x) * s + (v[3].y - y) * c 
    v[4].x, v[4].y = x + (v[4].x - x) * c - (v[4].y - y) * s, y + (v[4].x - x) * s + (v[4].y - y) * c 
end

local function buildHull(points) 
    if #points < 3 then return end
    
    local left, right = 1, 1
    for i = 2, #points do
        if points[left].x > points[i].x then 
            left = i 
        elseif points[left].x == points[i].x and points[i].y < points[left].y then
            left = i
        end
        
        if points[right].x < points[i].x then 
            right = i 
        elseif points[right].x == points[i].x and points[i].y > points[right].y then
            right = i
        end
    end
    
    local ax, ay, bx, by = points[left].x, points[left].y, points[right].x, points[right].y
    
    local up, down, s = {points[left], points[right]}, {points[left], points[right]}
    for i = 1, #points do
        s = online(points[i].x, points[i].y, ax, ay, bx, by)
        if i ~= left and i ~= right then
            if s <= 0 then
                up[#up + 1] = points[i]
            elseif s > 0 then
                down[#down + 1] = points[i]
            end
        end
    end
    
    table_sort(up, sort_up)
    
    local hull = {up[1], up[2]}
    
    if #up > 2 then
        for i = 3, #up do
            while #hull > 1 and not ccw(hull[#hull-1].x, hull[#hull-1].y, hull[#hull].x, hull[#hull].y, up[i].x, up[i].y) do
                table.remove(hull)
            end
            hull[#hull + 1] = up[i]
        end
    end
    
    table_sort(down, sort_down)
    
    if #down > 2 then
        local lim = #hull
        hull[#hull + 1] = down[2]
        for i = 3, #down do
            while #hull > lim and not ccw(hull[#hull-1].x, hull[#hull-1].y, hull[#hull].x, hull[#hull].y, down[i].x, down[i].y) do
                table.remove(hull)
            end
            hull[#hull + 1] = down[i]
        end
    end
    
    return hull
end


-- ROTATING BOX

local mt = {}

function mt:bbox(rb)
    local l, t, r, b, w, h = v_box(self.v)
    if rb then
        return l, t, r, b
    else
        return l, t, w, h
    end
end

function mt:move(dx, dy, world)
    self.x, self.y = self.x + dx, self.y + dy
    v_move(self.v, dx, dy)
    if world then
        world:modify(self)
    end
end

function mt:setPos(x, y, world)
    local dx, dy = x - self.x, y - self.y
    self.x, self.y = x, y
    v_move(self.v, dx, dy)
    if world then
        world:modify(self)
    end
end

function mt:getPos()
    return self.x, self.y
end

function mt:rotate(r, world)
    self.r = self.r + r
    v_rotate(self.v, self.x, self.y, r)
    if world then
        world:modify(self)
    end
end

function mt:contains(x, y)
    local l, t, r, b = v_box(self.v)
    
    if x >= l and x <= r and y >= t and y <= b then
        v_rotate(self.v, self.x, self.y, -self.r)
        
        x, y = rotate(x, y, -self.r, self.x, self.y)
        l, t, r, b = v_box(self.v)
        
        v_rotate(self.v, self.x, self.y, self.r)
        
        return x >= l and x <= r and y >= t and y <= b
    end
    return false
end


local lg = love.graphics
function mt:draw(dx, dy)
    dx, dy = dx or 0, dy or 0
    
    local v = self.v
    
    local r, g, b = lg.getColor()
    
    lg.setColor(r, g, b, 0.5)
    
    lg.polygon('fill', 
        v[1].x + dx, v[1].y + dy, 
        v[2].x + dx, v[2].y + dy, 
        v[3].x + dx, v[3].y + dy, 
        v[4].x + dx, v[4].y + dy
    )
    
    lg.setColor(r, g, b)
    
    lg.polygon('line', 
        v[1].x + dx, v[1].y + dy, 
        v[2].x + dx, v[2].y + dy, 
        v[3].x + dx, v[3].y + dy, 
        v[4].x + dx, v[4].y + dy
    )
    
    --lg.line(self.x + dx, self.y + dy, self.x + dx + cos(self.r) * self.hw, self.y + dy + sin(self.r) * self.hw)
end

function mt:onSide()
    local v = self.v
    return abs(v[1].x - v[2].x) <= 1 or abs(v[1].y - v[2].y) <= 1
end

local function RBox(cx, cy, w, h, angle)
    local hw, hh = 0.5 * w, 0.5 * h
    local v = {
        {x = cx - hw, y = cy - hh},
        {x = cx + hw, y = cy - hh},
        {x = cx + hw, y = cy + hh},
        {x = cx - hw, y = cy + hh}
    }
    local r = angle or 0

    if r ~= 0 then
        v_rotate(v, cx, cy, r)
    end
    
    return setmetatable({x = cx, y = cy, v = v, r = r, w = w, h = h, hw = 0.5 * w, hh = 0.5 * h}, {__index = mt})
end

local function MSum(item, other)
    local cx, cy = other:getPos()
    local ox, oy = item:getPos()
    
    local p, dx, dy, sx, sy = {}
    
    if abs(other.r - item.r) % (math.pi / 2) < 0.0000001 then
        -- build simple hull
        local m, d, index
        for i = 1, 4 do
            dx, dy = item.v[i].x - cx, item.v[i].y - cy
            
            sx, sy = item.v[i].x - ox, item.v[i].y - oy
            m, index = sx * (other.v[1].x + dx) + sy * (other.v[1].y + dy), 1
            
            for j = 2, 4 do
                d = sx * (other.v[j].x + dx) + sy * (other.v[j].y + dy)
                if d > m then m, index = d, j end
            end
            
            p[#p + 1] = {x = other.v[index].x + dx, y = other.v[index].y + dy}
        end
    else
        -- build convex hull            
        for i = 1, 4 do
            dx, dy = item.v[i].x - cx, item.v[i].y - cy
            for j = 1, 4 do
                p[#p + 1] = {x = other.v[j].x + dx, y = other.v[j].y + dy}
            end
        end
        
        p = buildHull(p)
    end
    
    return p
end

local function overlaps(item, other)
    local cx, cy = item:getPos()    
    local ox, oy = other:getPos()
    
    local p = MSum(other, item)
    
    -- if center of item is inside hull (p), then return TRUE 
    local j
    local h = {}
    for i = 1, #p do
        j = i < #p and i + 1 or 1
                    
        if side(cx, cy, ox, oy, p[i].x, p[i].y, p[j].x, p[j].y) < 0 then 
            return false -- (cx;cy) and (ox; oy) not in the same side of the edge
        end
    end
    
    return true
end

local function collides(item, other, goalX, goalY)
    local cx, cy = item:getPos()

    local ox, oy = other:getPos()
    
    local p = MSum(other, item)
    local dx, dy, sx, sy
    
    local j, cross
    local h = {}
    for i = 1, #p do
        j = i < #p and i + 1 or 1

        cross, sx, sy = segmentVsSegment(cx, cy, goalX, goalY, p[i].x, p[i].y, p[j].x, p[j].y)
        if cross then
            dx, dy = normalized(p[j].y - p[i].y, p[i].x - p[j].x)
            h[#h + 1] = {
                x = sx, y = sy, d = (sx - cx)^2 + (sy - cy)^2,
                nx = dx, ny = dy
            }
        end
    end
    
    if #h > 0 then
        if #h > 1 then
            table_sort(h, sort_d)
        end
        
        return {
            px = h[1].x + h[1].nx, py = h[1].y + h[1].ny, -- point of intersection with hull  = new position
            nx = h[1].nx, ny = h[1].ny,  -- normal
            dx = h[1].x + h[1].nx - cx, dy = h[1].y + h[1].ny - cy, -- correction move
            d = h[1].d
        }
    end
end


return {new = RBox, collides = collides, overlaps = overlaps, RBoxSum = MSum}